from twisted.conch.ssh import keys, userauth
from twisted.internet import defer


class AuthClient(userauth.SSHUserAuthClient):
    def __init__(self, options, *args):
        userauth.SSHUserAuthClient.__init__(self, str(options["user"]), *args)
        self.options = options

    def serviceStarted(self):
        userauth.SSHUserAuthClient.serviceStarted(self)

    def serviceStopped(self):
        pass

    def getPassword(self, prompt = None):
        return defer.succeed(str(self.options["password"]))

    def getPublicKey(self):
        return defer.succeed(keys.Key.fromString(self.options["privkey"]))

    def signData(self, publicKey, signData):
        return userauth.SSHUserAuthClient.signData(self, publicKey, signData)

    def getPrivateKey(self):
        return defer.succeed(keys.Key.fromString(self.options["privkey"]))



from zope.interface import Interface, Attribute, implements
from twisted.python.components import registerAdapter
from twisted.web.server import Session


from models import User

class IUserSession(Interface):
    user_id = Attribute("A user identifier.")


class UserSession(object):
    implements(IUserSession)

    def __init__(self, session):
        self.user_id = 0

    def get_user(self):
        if self.user_id:
            return User.get(self.user_id)
        else:
            return None

    def set_user(self, user):
        self.user_id = user.id


class LongSession(Session):
    sessionTimeout = 31536000

registerAdapter(UserSession, LongSession, IUserSession)





from twisted.internet.protocol import Protocol, Factory
from twisted.python import log

import json


class SockJSPubSubProtocol(Protocol):
    def connectionMade(self):
        self.factory.transports.add(self.transport)

    def connectionLost(self, reason):
        self.factory.transports.remove(self.transport)

    def dataReceived(self, data):
        json_data = json.loads(data)

        retval = {}
        ex = None

        try:
            func_retval = getattr(self, json_data["type"])(*json_data["args"])
        except Exception, e:
            log.err()
            ex = e.message
            func_retval = None

        if "callback" in json_data:
            retval["callback"] = json_data["callback"]
            retval["args"] = (ex, func_retval)
            self.transport.write(json.dumps(retval))

    def trigger(self, type, *args):
        raw_data = json.dumps({
            "type": type,
            "args": list(args),
        })
        print raw_data
        self.transport.write(raw_data)

    def trigger_all(self, type, *args):
        raw_data = json.dumps({
            "type": type,
            "args": list(args),
        })
        for transport in self.factory.transports:
            #if transport != self.transport:
            transport.write(raw_data)


class SockJSPubSubFactory(Factory):
    protocol = SockJSPubSubProtocol
    transports = set()

    def trigger_all(self, type, *args):
        raw_data = json.dumps({
            "type": type,
            "args": list(args),
        })

        for transport in self.transports:
            transport.write(raw_data)





from twisted.internet import reactor, task, defer
from twisted.conch.error import ConchError
from twisted.conch.ssh import forwarding
from twisted.conch.ssh import connection
from twisted.python import log

from webmux.session import SSHSession

conn = None

class SSHListenClientForwardingChannel(forwarding.SSHListenClientForwardingChannel):
    pass

class SSHConnectForwardingChannel(forwarding.SSHConnectForwardingChannel):
    pass

class KeepAlive(object):
    def __init__(self, conn):
        self.conn = conn
        self.globalTimeout = None
        self.lc = task.LoopingCall(self.sendGlobal)
        self.lc.start(300)

    def sendGlobal(self):
        d = self.conn.sendGlobalRequest("conch-keep-alive@twistedmatrix.com", "", wantReply = 1)
        d.addBoth(self._cbGlobal)
        self.globalTimeout = reactor.callLater(30, self._ebGlobal)

    def _cbGlobal(self, res):
        if self.globalTimeout:
            self.globalTimeout.cancel()
            self.globalTimeout = None

    def _ebGlobal(self):
        if self.globalTimeout:
            self.globalTimeout = None
            self.conn.transport.loseConnection()


class SSHConnection(connection.SSHConnection):
    def __init__(self, ssh, terminal, local_forward=None, remote_forward=None):
        self.ssh = ssh
        self.local_forward = local_forward
        self.remote_forward = remote_forward

        self.terminal = terminal
        connection.SSHConnection.__init__(self)

    def _beforeShutdown(self):
        for remotePort, hostport in self.remote_forward:
            log.msg("cancelling %s:%s" % (remotePort, hostport))
            self.cancelRemoteForwarding(remotePort)

    def _reConnect(self):
        self._beforeShutdown()
        self.transport.transport.loseConnection()

    def _do_localForwards(self):
        for localPort, hostport in self.local_forward:
            s = reactor.listenTCP(
                localPort,
                forwarding.SSHListenForwardingFactory(
                    conn,
                    hostport,
                    SSHListenClientForwardingChannel
                )
            )

            self.localForwards.append(s)

    def _do_remoteForwards(self):
        for remotePort, hostport in self.remote_forward:
            log.msg("asking for remote forwarding for %s:%s" %
                    (remotePort, hostport))
            self.requestRemoteForwarding(remotePort, hostport)
        reactor.addSystemEventTrigger("before", "shutdown", self._beforeShutdown)

    def processBacklog(self):
        if hasattr(self.transport, "sendIgnore"):
            KeepAlive(self)

        if self.local_forward:
            self._do_localForwards()

        if self.remote_forward:
            self._do_remoteForwards()

        self.openChannel(SSHSession(self.ssh, self, self.terminal.io))

    def serviceStarted(self):
        self.localForwards = []
        self.remoteForwards = {}
        if not isinstance(self, connection.SSHConnection):
            # make these fall through
            del self.__class__.requestRemoteForwarding
            del self.__class__.cancelRemoteForwarding

        self.processBacklog()

        self.ssh.connectionMade(self)

    def serviceStopped(self):
        lf = self.localForwards
        self.localForwards = []
        for s in lf:
            s.loseConnection()

        self.ssh.connectionLost(self)

    def requestRemoteForwarding(self, remotePort, hostport):
        data = forwarding.packGlobal_tcpip_forward(("0.0.0.0", remotePort))
        log.msg("requesting remote forwarding %s:%s" %(remotePort, hostport))
        try:
            yield self.sendGlobalRequest("tcpip-forward", data, wantReply=1)
        except:
            log.msg("remote forwarding %s:%s failed"%(remotePort, hostport))
            raise

        log.msg("accepted remote forwarding %s:%s" % (remotePort, hostport))
        self.remoteForwards[remotePort] = hostport
        log.msg(repr(self.remoteForwards))

    def cancelRemoteForwarding(self, remotePort):
        data = forwarding.packGlobal_tcpip_forward(("0.0.0.0", remotePort))
        self.sendGlobalRequest("cancel-tcpip-forward", data)
        log.msg("cancelling remote forwarding %s" % remotePort)
        try:
            del self.remoteForwards[remotePort]
        except:
            pass
        log.msg(repr(self.remoteForwards))

    def channel_forwarded_tcpip(self, windowSize, maxPacket, data):
        log.msg("%s %s" % ("FTCP", repr(data)))
        remoteHP, origHP = forwarding.unpackOpen_forwarded_tcpip(data)
        log.msg(self.remoteForwards)
        log.msg(remoteHP)
        if self.remoteForwards.has_key(remoteHP[1]):
            connectHP = self.remoteForwards[remoteHP[1]]
            log.msg("connect forwarding %s" % (connectHP,))
            return SSHConnectForwardingChannel(connectHP,
                                            remoteWindow = windowSize,
                                            remoteMaxPacket = maxPacket,
                                            conn = self)
        else:
            raise ConchError(connection.OPEN_CONNECT_FAILED, "don't know about that port")

    # def channelClosed(self, channel):
    #     self.terminal.close(channel)



from webmux.pubsub import SockJSPubSubProtocol, SockJSPubSubFactory
from webmux.models import Window, Terminal, Connection, User
from webmux import models

from twisted.python import log

SYNC_TABLES = {
    "connections": Connection,
    "windows": Window,
    "terminals": Terminal,
    "users": User,
}


class TerminalProtocol(SockJSPubSubProtocol):
    @staticmethod
    def data(terminal_id, data_to_write):
        Terminal.terminals[terminal_id].active_session.dataReceived(data_to_write.encode('utf8'))

    @staticmethod
    def write_to_terminal(terminal_id, data):
        Terminal.write_to_terminal(terminal_id, data)

    # def close_connection(self, terminal_id):
    #     self.sync("terminals", "delete", {"id": terminal_id})
    #     self.parent.delete_model(terminal_id, None)
    #     TerminalFactory.trigger_all("sync", "terminals", "delete", {"id": terminal_id})

    def sync(self, namespace, method, data):
        if method != "read" and not data:
            print "No data for sync: %s %s" % (namespace, method)
            return
        table = SYNC_TABLES[namespace]

        if method == "create":
            if "id" in data:
                raise Exception("Can't use self assigned ID!")

            data = table.create_model(data, self)

        elif method == "read":
            data = table.read_collection(self)

            log.msg("sync: %s %s %s" % (namespace, method, data))
            return data

        elif method == "update":
            if "id" not in data:
                raise Exception("No ID received, can't update model.")

            data = table.update_model(data, self)

        elif method == "delete":
            table.delete_model(data["id"], self)

        self.trigger_all("sync", namespace, method, data)

        log.msg("sync: %s %s %s" % (namespace, method, data))
        return data

TerminalFactory = SockJSPubSubFactory.forProtocol(TerminalProtocol)

# initialize the model with the list of connections so it would be able
# to send out events
models.factory = TerminalFactory




__author__ = 'Ron'

from webmux.terminal import Terminal

# TODO: merge SocketIO / StandardIO objects with Terminal objects
class SocketIO(object):
    def __init__(self, client):
        self.client = client

    def write(self, data):
        terminal_id = self.client.session.conn.terminal._id
        self.parent.write_to_terminal(terminal_id, data)

    def close(self, channel):
        terminal_id = self.client.session.conn.terminal._id
        self.parent.close_connection(terminal_id)

    def loseConnection(self):
        print "lose connection"

    def loseWriteConnection(self):
        print "lose write connection"


class WebTerminal(Terminal):
    def __init__(self, parent, id, cols, rows):
        self.parent = parent
        self._cols = cols
        self._rows = rows
        self._id = id
        self.active_session = None
        self.resize_callback = None

        self.io = SocketIO
        # TODO: fixme - merge SocketIO interface with Terminal interface
        self.io.parent = parent

    def get_window_size(self):
        return self._rows, self._cols, self._cols * 5, self._rows * 13

    def get_name(self):
        return "xterm-256color"

    def resize(self, cols, rows):
        self._cols = cols
        self._rows = rows
        self.resize_callback(rows, cols, cols * 5, rows * 13)

    def register_resize_callback(self, resize_callback):
        self.resize_callback = resize_callback

    def set_active_session(self, session):
        self.active_session = session




import sys

from twisted.internet import reactor
from twisted.python import log
from twisted.web import resource, server, static
from txsockjs.factory import SockJSResource

log.startLogging(sys.stdout)

from webmux.handlers import Signup, Login, Logout, Home
from webmux.protocols import TerminalFactory
from webmux.user import LongSession
from webmux.models import Terminal

from argparse import ArgumentParser

import webmux
import os

class StaticResource(resource.Resource):
    isLeaf = False


def init():
    for term in Terminal.select():
        term.connect()

def main(args):
    parser = ArgumentParser()
    parser.add_argument("-p", "--port", default=8080, type=int, help="Port to listen on.")

    args = parser.parse_args()

    WEBMUX_STATIC_PATH = os.path.join(webmux.__path__[0], "static")
    root = Home()

    static_path = resource.Resource()

    static_path.putChild("css", static.File(WEBMUX_STATIC_PATH + "/css"))
    static_path.putChild("js", static.File(WEBMUX_STATIC_PATH + "/js"))
    static_path.putChild("img", static.File(WEBMUX_STATIC_PATH + "/img"))

    root.putChild("terminal", SockJSResource(TerminalFactory))
    root.putChild("signup", Signup())
    root.putChild("login", Login())
    root.putChild("logout", Logout())
    root.putChild("static", static_path)
    site = server.Site(root)
    site.sessionFactory = LongSession

    reactor.listenTCP(args.port, site)

    reactor.callLater(0, init)

    reactor.run()




from twisted.conch.ssh import session, channel
from twisted.python import log
from twisted.conch.ssh import connection
from twisted.internet import reactor


import time

class SSHSession(channel.SSHChannel):
    name = 'session'
    def __init__(self, ssh, conn, io_handler):
        channel.SSHChannel.__init__(self)
        self.ssh = ssh
        self.conn = conn
        self.stdio = None
        self.io_handler = io_handler
        self.window_size = None
        self.last_update = None
        self.queue_call = False

    def allocatePty(self):
        term = self.conn.terminal.get_name()
        log.msg('terminal name:', term)

        self.window_size = self.conn.terminal.get_window_size()
        log.msg('terminal window size:', self.window_size)

        ptyReqData = session.packRequest_pty_req(term, self.window_size, '')
        self.last_update = time.time()
        self.conn.sendRequest(self, 'pty-req', ptyReqData)

        self.conn.terminal.register_resize_callback(self._windowResized)

    def channelOpen(self, foo):
        log.msg('session %s open' % self.id)
        self.conn.terminal.enter()

        c = session.SSHSessionClient()

        # TODO: figure out if this is right
        c.dataReceived = self.write

        c.connectionLost = self.sendEOF
        c.session = self

        self.conn.terminal.set_active_session(c)

        self.stdio = self.io_handler(c)

        self.allocatePty()
        self.conn.sendRequest(self, 'shell', '')

    def dataReceived(self, data):
        self.stdio.write(data)

    def extReceived(self, t, data):
        if t==connection.EXTENDED_DATA_STDERR:
            log.msg('got %s stderr data' % len(data))
            self.stdio.write(data)

    def eofReceived(self):
        log.msg('got eof')
        self.stdio.loseWriteConnection()

    def closeReceived(self):
        log.msg('remote side closed %s' % self)
        self.stdio.close(self)
        self.conn.sendClose(self)

    def closed(self):
        log.msg('closed %s' % self)
        self.stdio.close(self)
        log.msg(repr(self.conn.channels))

    def sendEOF(self, *args):
        log.msg('sendEOF')
        self.conn.sendEOF(self)

    def stopWriting(self):
        self.stdio.pauseProducing()

    def startWriting(self):
        self.stdio.resumeProducing()

    def _windowResized(self, *args):
        self.window_size = self.conn.terminal.get_window_size()
        self.updateWindowSize()

    def updateWindowSize(self):
        time_diff = float(time.time() - self.last_update)
        if time_diff < 1 and not self.queue_call:
            reactor.callLater(time_diff, self.updateWindowSize)
            self.queue_call = True
            return

        self.queue_call = False
        self.last_update = time.time()
        self.conn.sendRequest(self, 'window-change', session.packRequest_window_change(self.window_size))




__author__ = 'Ron'

class Terminal(object):
    def get_window_size(self):
        raise NotImplementedError()

    def get_name(self):
        raise NotImplementedError()

    def enter(self):
        pass

    def leave(self):
        pass






from sys import stdout

from twisted.python.log import startLogging, err

from twisted.internet import reactor
from twisted.internet.defer import Deferred

from twisted.conch.ssh.common import NS
from twisted.conch.scripts.cftp import ClientOptions
from twisted.conch.ssh.filetransfer import FileTransferClient, FXF_WRITE, FXF_CREAT, FXF_TRUNC
from twisted.conch.client.connect import connect
from twisted.conch.client.default import SSHUserAuthClient, verifyHostKey
from twisted.conch.ssh.connection import SSHConnection
from twisted.conch.ssh.channel import SSHChannel


class SFTPSession(SSHChannel):
    name = 'session'

    def channelOpen(self, whatever):
        d = self.conn.sendRequest(
            self, 'subsystem', NS('sftp'), wantReply=True)
        d.addCallbacks(self._cbSFTP)


    def _cbSFTP(self, result):
        client = FileTransferClient()
        client.makeConnection(self)
        self.dataReceived = client.dataReceived
        self.conn._sftp.callback(client)



class SFTPConnection(SSHConnection):
    def serviceStarted(self):
        self.openChannel(SFTPSession())


def sftp(user, host, port):
    options = ClientOptions()
    options['host'] = host
    options['port'] = port
    options.identitys = ['~/.ssh/amazon.pem']

    conn = SFTPConnection()
    conn._sftp = Deferred()
    auth = SSHUserAuthClient(user, options, conn)
    connect(host, port, options, verifyHostKey, auth)
    return conn._sftp


def transfer(client):
    d = client.openFile('/home/ec2-user/test.txt', FXF_WRITE | FXF_CREAT | FXF_TRUNC, {})
    def cbWrite(f):
        print 'File created'
        f.writeChunk(5, 'hello')

    d.addCallback(cbWrite)
    return d


def main():
    startLogging(stdout)

    user = 'ec2-user'
    host = 'crosswiselabs.com'
    port = 22
    d = sftp(user, host, port)
    d.addCallback(transfer)
    d.addErrback(err, "Problem with SFTP transfer")
    d.addCallback(lambda ignored: reactor.stop())
    reactor.run()


if __name__ == '__main__':
    main()


#!/usr/bin/env python

# to distribute:
# python setup.py register sdist upload

from setuptools import setup
from pip.req import parse_requirements

import os
import webmux

reqs = []

try:
    install_reqs = parse_requirements(os.path.join(os.path.dirname(__file__), 'requirements.txt'))

    reqs = [str(ir.req) for ir in install_reqs]

except:
    pass

setup(
    name='webmux',
    version=webmux.__version__,
    description='An open-source web based SSH terminal multiplexer',
    long_description='An open-source web based SSH terminal multiplexer',
    author='Ron Reiter (@ronreiter)',
    author_email='ron.reiter@gmail.com',
    url='http://github.com/ronreiter/webmux',
    packages=['webmux'],
    scripts=['webmuxd'],
    license='MIT',

    # TODO: how do you use requirements.txt without moving it into the package?
    install_requires=[
        'Twisted==13.2.0',
        'cffi==0.8.1',
        'cryptography==0.2.1',
        'futures==2.1.6',
        'parse==1.4.1',
        'pyOpenSSL==0.14',
        'pyasn1==0.1.7',
        'pycparser==2.10',
        'pycrypto==2.6.1',
        'six==1.5.2',
        'stuf==0.9.4',
        'txsockjs==1.2.1',
        'wsgiref==0.1.2',
        'zope.interface==4.1.0',
        'SQLObject==1.5.1',
        'Jinja2==2.7.2',
    ]
)



from twisted.web.util import redirectTo
from webmux.user import IUserSession
from webmux.models import User
from twisted.web import resource

import webmux
import os

from jinja2 import Environment, FileSystemLoader
jinja = Environment(loader=FileSystemLoader(os.path.join(webmux.__path__[0], 'templates')))

from Crypto.Cipher import AES

secret = 'ca621b281350e4ebe6d92629b299fb84'
init_vector = 'IV'
COOKIE_NAME = 'session'

class SecureCookieResource(resource.Resource):
    def get_session(self, request):
        cookie_data = ""
        request.getCookie(COOKIE_NAME)
        request.addCookie(COOKIE_NAME, cookie_data, path='/', max_age=60*60*24*365)

        obj = AES.new(secret, AES.MODE_CBC, init_vector)
        message = "The answer is no"
        ciphertext = obj.encrypt(message)

        obj2 = AES.new(secret, AES.MODE_CBC, init_vector)
        obj2.decrypt(ciphertext)


class Home(SecureCookieResource):
    isLeaf = False

    def getChild(self, name, request):
        if name == "":
            return self
        return resource.Resource.getChild(self, name, request)

    def render_GET(self, request):
        user_session = IUserSession(request.getSession())

        user = user_session.get_user()
        if user:
            #user_session.email = "foo"
            return jinja.get_template("index.html").render(user=user).encode("utf8")
        else:
            num_users = User.select().count()
            if num_users == 0:
                return redirectTo("signup", request)
            else:
                return redirectTo("login", request)


class Signup(SecureCookieResource):
    isLeaf = True

    def render_GET(self, request):
        num_users = User.select().count()
        session = IUserSession(request.getSession())

        return jinja.get_template("signup.html").render(
            session=session,
            create_admin=num_users == 0
        ).encode("utf8")

    def render_POST(self, request):
        session = IUserSession(request.getSession())

        if request.args["password"][0] != request.args["confirm"][0]:
            return redirectTo("signup", request)

        user = User.signup(request.args["email"][0], request.args["password"][0])

        if user:
            session.set_user(user)
        else:
            return redirectTo("signup", request)

        return redirectTo("/", request)


class Login(SecureCookieResource):
    isLeaf = True

    def render_GET(self, request):
        session = IUserSession(request.getSession())

        return jinja.get_template("login.html").render(session=session).encode("utf8")

    def render_POST(self, request):
        session = IUserSession(request.getSession())

        user = User.login(request.args["email"][0], request.args["password"][0])
        if user:
            session.set_user(user)
        else:
            return redirectTo("login", request)

        return redirectTo("/", request)


class Logout(SecureCookieResource):
    isLeaf = True

    def render_GET(self, request):
        request.getSession().expire()

        return redirectTo("login", request)



from twisted.conch.client import direct
from twisted.internet import defer, reactor, endpoints
from twisted.conch.ssh.transport import SSHClientTransport, SSHCiphers

from webmux.connection import SSHConnection
from webmux.auth import AuthClient
from twisted.python import log


class SSH(object):
    def __init__(self, user, password, host, port, privkey):

        self.options = {
            "user": user,
            "password": password,
            "host": host,
            "port": port,
            "privkey": privkey,
            "identity": None,
            "ciphers": SSHCiphers.cipherMap.keys(),
            "macs": SSHCiphers.macMap.keys(),
            "host-key-algorithms": SSHClientTransport.supportedPublicKeys,
            "known-hosts": None,
            "user-authentication": ["publickey", "password", "keyboard-interactive"],
            "logfile": None,
            "option": None,
            "compress": None,
            "version": None,
            "log": None,
            "nox11": None,
            "agent": None,
            "noagent": None,
            "reconnect": None,
        }

    @defer.inlineCallbacks
    def connect(self, terminal=None, on_error=None):
        connection_string = "tcp:host={host}:port={port}".format(
            host=self.options["host"],
            port=self.options["port"]
        )

        log.msg("Connecting with connection string %s" % connection_string)

        # TODO: add port forwarding here
        ssh_connection = SSHConnection(self, terminal)

        # TODO: verify host keys
        vhk = lambda *a: defer.succeed(1)

        # Build a basic auth client. There's a Twisted implementation
        # for this, but we need something simpler that does not access
        # keys that are stored on the disk
        uao = AuthClient(self.options, ssh_connection)

        d = defer.Deferred()
        factory = direct.SSHClientFactory(d, self.options, vhk, uao)
        if on_error:
            d.addErrback(on_error)

        endpoint = endpoints.clientFromString(reactor, connection_string)

        wp = None

        try:
            wp = yield endpoint.connect(factory)
        except Exception:
            terminal.leave()

        defer.returnValue(wp)

    def connectionMade(self, conn):
        pass

    def connectionLost(self, conn):
        conn.terminal.leave()





from sqlobject import *
from sqlobject.inheritance import InheritableSQLObject

from webmux.ssh import SSH
from webmux.web_terminal import WebTerminal

from twisted.python import log

import hashlib


sqlhub.processConnection = connectionForURI('sqlite:webmux.db')
factory = None

MAX_TERM_HISTORY = 10000


class User(SQLObject):
    email = StringCol()
    password = StringCol()
    is_admin = BoolCol()
    #desktops = RelatedJoin('Desktop', joinColumn='user', otherColumn='desktop', intermediateTable='desktop_users')

    @classmethod
    def login(cls, email, password):
        users = User.selectBy(email=email, password=hashlib.md5(password).hexdigest())
        if users.count():
            return users[0]

    @classmethod
    def signup(cls, email, password):
        num_users = User.select().count()

        users = cls.selectBy(email=email)

        if users.count():
            raise Exception("Can't sign up, user already exists")

        is_admin = True if num_users == 0 else False
        new_user = User(
            email=email,
            password=hashlib.md5(password).hexdigest(),
            is_admin=is_admin
        )

        return new_user

    def serialize(self):
        return {
            "id": self.id,
            "email": self.email,
            "password": self.password,
            "is_admin": self.is_admin,
        }

    @classmethod
    def read_collection(cls, protocol):
        # TODO: check if we are an admin

        return [x.serialize() for x in cls.select()]

    @classmethod
    def create_model(cls, data, protocol):
        # TODO: check if we are an admin

        user = User(
            email=data["email"],
            password=data["password"],
            is_admin=data["is_admin"]
        )

        return user.serialize()

    @classmethod
    def update_model(cls, data, protocol):
        # TODO: check if we are an admin

        model = cls.get(data["id"])
        model.email = data["email"]
        model.password = data["password"]
        model.is_admin = data["is_admin"]

        return model.serialize()

    @classmethod
    def delete_model(cls, user_id, protocol):
        # TODO: check if we are an admin

        User.delete(user_id)
        return True

class Connection(SQLObject):
    name = StringCol()
    user = StringCol()
    password = StringCol()
    host = StringCol()
    port = IntCol()
    privkey = StringCol()

    def serialize(self):
        return {
            "id": self.id,
            "name": self.name,
            "user": self.user,
            "password": self.password,
            "host": self.host,
            "port": self.port,
            "privkey": self.privkey,
        }

    @classmethod
    def read_collection(cls, protocol):
        return [x.serialize() for x in cls.select()]

    @classmethod
    def create_model(cls, data, protocol):
        if not data["user"]:
            raise Exception("No user provided!")

        if not data["privkey"] and not data["password"]:
            raise Exception("No authentication method!")

        model = Connection(
            name=data["name"],
            user=data["user"],
            password=data["password"],
            host=data["host"],
            port=int(data["port"]),
            privkey=data["privkey"],
        )
        return model.serialize()

    @classmethod
    def update_model(cls, data, protocol):
        model = cls.get(data["id"])
        model.name = data["name"]
        model.user = data["user"]
        model.password = data["password"]
        model.host = data["host"]
        model.port = data["port"]
        model.privkey = data["privkey"]

    @classmethod
    def delete_model(cls, connection_id, protocol):
        if Terminal.select(connection=connection_id).count():
            raise Exception("can't remove a connection if it is open")

        Connection.delete(connection_id)
        return True

class Window(SQLObject):
    rows = IntCol()
    cols = IntCol()
    top = IntCol()
    left = IntCol()
    hidden = BoolCol()
    z_index = IntCol()
    #desktop = ForeignKey("Desktop")

    def serialize(self):
        return {
            "id": self.id,
            "rows": self.rows,
            "cols": self.cols,
            "top": self.top,
            "left": self.left,
            "hidden": self.hidden,
            "z_index": self.z_index,
        }

    @classmethod
    def read_collection(cls, protocol):
        return [x.serialize() for x in cls.select()]

    @classmethod
    def create_model(cls, data, protocol):
        log.msg("create window %s" % data)
        last_z_window = None
        for win in Window.select(orderBy=DESC(Window.q.z_index)):
            last_z_window = win
            break

        if not last_z_window:
            last_z_index = 1
        else:
            last_z_index = last_z_window.z_index + 1

        model = Window(
            rows=data.get("rows", 25),
            cols=data.get("cols", 80),
            top=data.get("top", 100),
            left=data.get("left", 20),
            hidden=data.get("hidden", False),
            z_index=data.get("z_index", last_z_index)
        )

        if not model.left:
            model.left = 20 * (cls.select().count() % 5 + 1)

        if not model.top:
            model.top = 50 * (cls.select().count() % 5 + 1)

        return model.serialize()

    @classmethod
    def update_model(cls, data, protocol):
        log.msg("update window %s" % data)

        model = cls.get(data["id"])
        has_changed = model.cols != data["cols"] or model.rows != data["rows"]

        model.cols = data["cols"]
        model.rows = data["rows"]
        model.left = data["left"]
        model.top = data["top"]
        model.hidden = data["hidden"]
        model.z_index = data["z_index"]

        if model.cols < 20:
            model.cols = 20
        if model.cols > 1000:
            model.cols = 1000
        if model.rows < 5:
            model.rows = 5
        if model.rows > 500:
            model.rows = 500
        if model.left < 0:
            model.left = 0
        if model.top < 50:
            model.top = 50
        if model.z_index < 0:
            model.z_index = 0

        if has_changed:
            for term in Terminal.selectBy(window=model):
                Terminal.terminals[term.id].resize(model.cols, model.rows)
                term.history = str()

        return model.serialize()

    @classmethod
    def delete_model(cls, model_id, protocol):
        model = Window.get(model_id)

        if not model:
            return

        log.msg("delete window %s" % model)

        for term in Terminal.selectBy(window=model):
            protocol.sync("terminals", "delete", {"id": term.id})

        Window.delete(model.id)
        return True


class Terminal(SQLObject):
    history = StringCol()
    window = ForeignKey("Window")
    conn = ForeignKey("Connection")
    terminals = {}
    factory = None

    def serialize(self):
        return {
            "id": self.id,
            "history": self.history,
            "window_id": self.window.id,
            "connection_id": self.conn.id,
        }

    @classmethod
    def read_collection(cls, protocol):
        return [x.serialize() for x in cls.select()]

    @classmethod
    def create_model(cls, data, protocol):
        conn = Connection.get(data["connection_id"])
        win = Window.get(data["window_id"])

        model = Terminal(
            history="",
            window=win,
            conn=conn,
        )

        model.connect()

        return model.serialize()

    def connect(self):
        log.msg("connecting to %s with user %s" % (self.conn.host, self.conn.user))
        self.history = str("Connecting to %s...\r\n" % self.conn.host)

        if not self.conn.user:
            raise Exception("No user given.")
        if not self.conn.password and not self.conn.privkey:
            raise Exception("No authentication details given.")

        web_term = WebTerminal(self, self.id, self.window.cols, self.window.rows)

        Terminal.terminals[self.id] = web_term

        ssh = SSH(
            user=self.conn.user,
            password=self.conn.password,
            host=self.conn.host,
            port=self.conn.port,
            privkey=self.conn.privkey,
        )

        def err(e):
            self.write_to_terminal(self.id, e.value[0])

        ssh.connect(web_term, on_error=err)


    @classmethod
    def update_model(cls, data, protocol):
        model = cls.get(data["id"])
        return model.serialize()

    @classmethod
    def delete_model(cls, terminal_id, protocol):
        log.msg("delete terminal %s" % terminal_id)

        model = Terminal.get(terminal_id)
        if not model:
            return False

        window_id = model.window.id

        Terminal.delete(terminal_id)

        kill_window = True
        for terminal in Terminal.select():
            if terminal.window.id == window_id:
                kill_window = False

        if kill_window:
            Window.delete(window_id)

        return True

    def write_to_terminal(self, terminal_id, data):
        #log.msg("write to terminal %s %s" % (terminal_id, data))
        model = Terminal.get(terminal_id)

        model.history += str(data)
        if len(model.history) > MAX_TERM_HISTORY:
            model.history = model.history[-MAX_TERM_HISTORY:]

        factory.trigger_all("data", terminal_id, str(data))

    # TODO: why is this not static?
    def close_connection(self, terminal_id):
        Terminal.delete(terminal_id)
        factory.trigger_all("sync", "terminals", "delete", {"id": terminal_id})

# class Desktop(SQLObject):
#     name = StringCol()
#     short_link = StringCol()
#     users = RelatedJoin('User', joinColumn='desktop', otherColumn='user', intermediateTable='desktop_users')


class Preference(SQLObject):
    key = StringCol(alternateID=True)
    value = StringCol()


User.createTable(ifNotExists=True)
Connection.createTable(ifNotExists=True)
#Desktop.createTable(ifNotExists=True)
Preference.createTable(ifNotExists=True)
Window.createTable(ifNotExists=True)
Terminal.createTable(ifNotExists=True)
